from tempfile import TemporaryDirectory
from unittest import TestCase

from .storage import InvoiceStorage


class InvoiceStorageTest(TestCase):
    def test_list(self):
        with TemporaryDirectory() as testdir:
            storage = InvoiceStorage(testdir)
            assert len(list(storage.list())) == 0

    def test_contact_invoice(self):
        with TemporaryDirectory() as testdir:
            storage = InvoiceStorage(testdir)

            # Create bank
            storage.update_bank("EUR", bank="Test")

            # Create contact
            storage.update_contact(
                "test",
                "Name",
                "Address",
                "City",
                "Country",
                "noreply@example.com",
                "",
                "",
                "EUR",
                "test",
            )
            assert storage.read_contact("test")["name"] == "Name"

            # Create invoice
            storage.create("test", rate="100", item="Test item")

            # List it
            invoices = list(storage.list())
            assert len(invoices) == 1



from __future__ import annotations

import datetime
import os
import re
from configparser import RawConfigParser
from glob import glob

import jinja2
from django.utils.functional import cached_property
from filelock import FileLock

from .invoices import Invoice, Proforma, Quote

LATEX_SUBS = (
    (re.compile(r"\\"), r"\\textbackslash"),
    (re.compile(r"([{}_#%&$])"), r"\\\1"),
    (re.compile(r"~"), r"\~{}"),
    (re.compile(r"\^"), r"\^{}"),
    (re.compile(r'"'), r"''"),
    (re.compile(r"\.\.\.+"), r"\\ldots"),
)

DASH_RE = re.compile(r"-")


def escape_tex(value):
    newval = value
    for pattern, replacement in LATEX_SUBS:
        newval = pattern.sub(replacement, newval)
    return newval


def escape_dash(value):
    return DASH_RE.sub(r"--", value)


class InvoiceStorage:
    data = "data"
    pdf = "pdf"
    tex = "tex"
    config = "config"
    contacts = "contacts"
    banks = "banks"
    default_due = 15

    template = "{year}{month}{order}.ini"
    order = "{:02d}"

    base = Invoice

    def __init__(self, basedir="."):
        self.basedir = basedir
        lockfile = self.path(self.config, "lock")
        self.ensure_dir(lockfile)
        self.lock = FileLock(lockfile)
        self.jinja = jinja2.Environment(
            block_start_string=r"\BLOCK{",
            block_end_string="}",
            variable_start_string=r"\VAR{",
            variable_end_string="}",
            comment_start_string=r"\#{",
            comment_end_string="}",
            line_statement_prefix="%%",
            line_comment_prefix="%#",
            trim_blocks=True,
            autoescape=False,
            loader=jinja2.FileSystemLoader(os.path.abspath(basedir)),
        )
        self.jinja.filters["escape_tex"] = escape_tex
        self.jinja.filters["escape_dash"] = escape_dash

    @staticmethod
    def ensure_dir(filename):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def path(self, *args):
        return os.path.join(self.basedir, *args)

    def glob(self, year=None, month=None):
        if year:
            full_year = str(year)
            year = f"{year % 2000:02d}"
        else:
            full_year = "[0-9][0-9][0-9][0-9]"
            year = "[0-9][0-9]"
        month = f"{month:02d}" if month else "[0-9][0-9]"
        mask = self.template.format(
            year=year,
            month=month,
            order="*",
            full_year=full_year,
        )
        return sorted(glob(self.path(self.data, mask)))

    def list(self, year=None, month=None):
        for filename in self.glob(year, month):
            yield self.base(self, filename)

    def get(self, invoice):
        if "/" not in invoice:
            return self.base(self, self.path(self.data, f"{invoice}.ini"))
        return self.base(self, self.path(invoice))

    @cached_property
    def settings(self):
        data = RawConfigParser()
        data.read(self.path(self.config, "config.ini"))
        return dict(data["config"])

    def find_filename(self):
        today = datetime.date.today()
        year = today.strftime("%y")
        full_year = today.strftime("%Y")
        month = today.strftime("%m")
        for i in range(1, 1000):
            filename = self.path(
                self.data,
                self.template.format(
                    year=year,
                    month=month,
                    order=self.order.format(i),
                    full_year=full_year,
                ),
            )
            if os.path.exists(filename):
                continue
            return filename
        raise ValueError("Failed to find invoice number!")

    def create(self, contact, duedelta: int | None = None, **kwargs):
        if duedelta is None:
            duedelta = self.default_due
        with self.lock:
            today = datetime.date.today()
            due = today + datetime.timedelta(days=duedelta)
            filename = self.find_filename()
            invoice = RawConfigParser()
            invoice.add_section("invoice")
            invoice.set("invoice", "contact", contact)
            invoice.set("invoice", "date", today.isoformat())
            invoice.set("invoice", "due", due.isoformat())
            # Apply defaults from contact
            contact = self.read_contact(contact)
            for key, value in contact.items():
                if not key.startswith("default_"):
                    continue
                invoice.set("invoice", key[8:], value)
            # Apply passed value
            for key, value in kwargs.items():
                invoice.set("invoice", key, value)
            # Ensure rate and item are present
            for key in ("rate", "item"):
                if not invoice.has_option("invoice", key):
                    invoice.set("invoice", key, "")
            # Store the file
            self.ensure_dir(filename)
            with open(filename, "w") as handle:
                invoice.write(handle)
            return filename

    def contact_path(self, name):
        return self.path(self.contacts, f"{name}.ini")

    def parse_contact(self, name):
        data = RawConfigParser()
        data.read(self.contact_path(name))
        return data

    def read_contact(self, name):
        data = self.parse_contact(name)
        try:
            return dict(data["contact"])
        except KeyError as error:
            raise ValueError(f"Contact {name} not found!") from error

    def bank_path(self, name):
        return self.path(self.banks, f"{name}.ini")

    def read_bank(self, name, extra_suffix=None):
        data = RawConfigParser()
        data.read(self.bank_path(name))
        if extra_suffix:
            data.read(self.path(self.banks, f"{name}-{extra_suffix}.ini"))
        try:
            return dict(data["bank"])
        except KeyError as error:
            raise ValueError(f"Bank account {name} not found!") from error

    def update_bank(self, name, **kwargs):
        filename = self.bank_path(name)
        self.ensure_dir(filename)
        if os.path.exists(filename):
            bank = RawConfigParser()
            bank.read(filename)
        else:
            bank = RawConfigParser()

        if not bank.has_section("bank"):
            bank.add_section("bank")

        for key, value in kwargs.items():
            bank.set("bank", key, value)

        with open(filename, "w") as handle:
            bank.write(handle)
        return filename

    def update_contact(  # noqa: PLR0913
        self,
        key,
        name,
        address,
        city,
        country,
        email,
        tax_reg,
        vat_reg,
        default_currency,
        default_category,
    ):
        filename = self.contact_path(key)
        self.ensure_dir(filename)
        if os.path.exists(filename):
            contact = self.parse_contact(key)
        else:
            contact = RawConfigParser()

        if not contact.has_section("contact"):
            contact.add_section("contact")

        contact.set("contact", "name", name)
        contact.set("contact", "address", address)
        contact.set("contact", "city", city)
        contact.set("contact", "country", country)
        contact.set("contact", "email", email)
        contact.set("contact", "tax_reg", tax_reg)
        contact.set("contact", "vat_reg", vat_reg)
        contact.set("contact", "default_currency", default_currency)
        contact.set("contact", "default_category", default_category)

        with open(filename, "w") as handle:
            contact.write(handle)
        return filename


class QuoteStorage(InvoiceStorage):
    data = "quotes"
    pdf = "quotes"
    tex = "quotes"
    template = "Q{full_year}{order}.ini"
    default_due = 30

    base = Quote


class WebStorage(InvoiceStorage):
    template = "W{year}{month}{order}.ini"
    order = "{:03d}"


class ProformaStorage(InvoiceStorage):
    data = "proforma"
    pdf = "proforma"
    tex = "proforma"
    template = "P{full_year}{order}.ini"
    order = "{:05d}"

    base = Proforma



import json
import os
from typing import ClassVar
from urllib.request import urlopen

from fakturace.data import CACHE_DIR, RATE_URL


class Rates:
    datacache: ClassVar[dict[str, float]] = {}

    @classmethod
    def download(cls, date: str):
        if not os.path.exists(CACHE_DIR):
            os.makedirs(CACHE_DIR)
        cache_file = os.path.join(CACHE_DIR, f"rates-{date}")

        # Filesystem cache
        if date not in cls.datacache and os.path.exists(cache_file):
            with open(cache_file) as handle:
                cls.datacache[date] = json.load(handle)

        # Load remotely
        if date not in cls.datacache:
            cls.datacache[date] = {}
            parts = date.split("-")
            handle = urlopen(RATE_URL.format(*parts))
            content = handle.read().decode("utf-8")
            for line in content.splitlines():
                if "|" not in line:
                    continue
                parts = line.split("|")
                if parts[4] in ("kurz", "Rate"):
                    continue
                cls.datacache[date][parts[3]] = float(parts[4].replace(",", "."))

            # Update filesystem cache
            with open(cache_file, "w") as handle:
                json.dump(cls.datacache[date], handle)

        return cls.datacache[date]

    @classmethod
    def get(cls, date: str, currency: str):
        if currency == "CZK":
            return 1
        rates = cls.download(date)
        return rates[currency]



import os

DEFAULTS = {
    "quantity": "1",
    "currency": "EUR",
    "bank_suffix": "",
    "currency_note": "",
    "note": "",
    "category": "none",
    "template": "template/invoice.tex",
    "row": "template/row.tex",
    "vat": "0",
}
CONTACT = {"vat_reg": "", "tax_reg": ""}
RATE_URL = (
    "http://www.cnb.cz/cs/financni_trhy/devizovy_trh/"
    "kurzy_devizoveho_trhu/denni_kurz.txt?date={2}.{1}.{0}"
)
CACHE_DIR = os.path.expanduser("~/.cache/fakturace")



import datetime
import subprocess
from argparse import ArgumentParser
from fnmatch import fnmatch
from xml.etree import ElementTree

from vies.types import VATIN

from .storage import InvoiceStorage, ProformaStorage, QuoteStorage, WebStorage

COMMANDS = {}


def register_command(command):
    """Register a command in command line interface."""
    COMMANDS[command.__name__.lower()] = command
    return command


class Command:
    """Basic command object."""

    def __init__(self, args):
        """Construct Command object."""
        self.args = args
        if args.quotes:
            self.storage = QuoteStorage()
        elif args.web:
            self.storage = WebStorage()
        elif args.proforma:
            self.storage = ProformaStorage()
        else:
            self.storage = InvoiceStorage()

    @classmethod
    def add_parser(cls, subparser):
        """Create parser for command line."""
        return subparser.add_parser(cls.__name__.lower(), description=cls.__doc__)

    def run(self):
        """Execute the command."""
        raise NotImplementedError


class FilterCommand(Command):
    """List invoices."""

    @classmethod
    def add_parser(cls, subparser):
        """Create parser for command line."""
        parser = super().add_parser(subparser)
        parser.add_argument(
            "--year",
            type=int,
            help="Year to process",
            default=datetime.date.today().year,
        )
        parser.add_argument(
            "--filter",
            help="Filter by ID",
        )
        parser.add_argument(
            "--vat",
            action="store_true",
            help="Include VAT",
            default=False,
        )
        parser.add_argument("match", nargs="?", help="Match string to find")
        return parser

    def match(self, invoice):
        if self.args.filter and not fnmatch(invoice.invoiceid, self.args.filter):
            return False

        if not self.args.match:
            return True
        match = self.args.match.lower()
        return (
            match in invoice.invoice["item"].lower()
            or match in invoice.invoice["contact"].lower()
            or match in invoice.contact["name"].lower()
        )

    def list(self):
        for invoice in self.storage.list(self.args.year):
            if self.match(invoice):
                yield invoice


@register_command
class List(FilterCommand):
    """List invoices."""

    def run(self):
        """Execute the command."""
        total = 0
        for invoice in self.list():
            amount = invoice.amount_czk_vat if self.args.vat else invoice.amount_czk
            print(
                "{}: {} {} ({:.2f} CZK): {} [{}]".format(
                    invoice.invoiceid,
                    invoice.amount,
                    invoice.currency,
                    amount,
                    invoice.invoice["item"],
                    invoice.contact["name"],
                ),
            )
            total += amount
        print()
        print(f"Total: {total:.2f} CZK")


@register_command
class XMLExport(FilterCommand):
    """XML exportinvoices."""

    def add_element(self, root, name: str, text: str | None = None):
        added = ElementTree.SubElement(root, name)
        if text is not None:
            added.text = text
        return added

    def add_amounts(self, root, invoice, prefix: str = ""):
        dph = self.add_element(root, "SouhrnDPH")
        vat_rate = int(invoice.invoice["vat"])
        fixed_rates = {0, 5, 22}
        if vat_rate in fixed_rates:
            self.add_element(
                dph, f"Zaklad{vat_rate}", invoice.invoice[f"{prefix}total"]
            )
            if vat_rate > 0:
                self.add_element(
                    dph, f"DPH{vat_rate}", invoice.invoice[f"{prefix}total_vat"]
                )
            fixed_rates.remove(vat_rate)
            for rate in fixed_rates:
                self.add_element(dph, f"Zaklad{rate}", "0")
            for rate in fixed_rates:
                if rate > 0:
                    self.add_element(dph, f"DPH{rate}", "0")
        else:
            dalsi = self.add_element(dph, "SeznamDalsiSazby")
            sazba = self.add_element(dalsi, "DalsiSazba")
            self.add_element(sazba, "Sazba", invoice.invoice["vat"])
            self.add_element(sazba, "Zaklad", invoice.invoice[f"{prefix}total"])
            self.add_element(sazba, "DPH", invoice.invoice[f"{prefix}total_vat"])
        self.add_element(root, "Celkem", invoice.invoice[f"{prefix}total_sum"])

    def run(self):  # noqa: PLR0915
        """Execute the command."""
        document = ElementTree.Element("MoneyData")
        invoices = ElementTree.SubElement(document, "SeznamFaktVyd")

        for invoice in self.list():
            output = ElementTree.SubElement(invoices, "FaktVyd")
            self.add_element(output, "Doklad", invoice.invoiceid)
            self.add_element(output, "CisRada", "0")
            self.add_element(output, "Popis", invoice.invoice["item"])
            self.add_element(output, "Vystaveno", invoice.invoice["date"])
            self.add_element(output, "DatUcPr", invoice.invoice["date"])
            self.add_element(output, "PlnenoDPH", invoice.invoice["date"])
            self.add_element(output, "Splatno", invoice.invoice["due"])
            self.add_element(output, "DatSkPoh", invoice.invoice["date"])
            self.add_element(output, "KodDPH", "19Ř21")
            self.add_element(output, "ZjednD", "0")
            self.add_element(output, "VarSymbol", invoice.invoiceid)

            # Druh (N: normální, L: zálohová, F: proforma, D: doklad k přijaté platbě)
            self.add_element(output, "Druh", "N")
            self.add_element(
                output,
                "Dobropis",
                "0" if float(invoice.invoice["total_sum"]) > 0 else "1",
            )
            self.add_element(output, "ZpVypDPH", "1")
            self.add_element(output, "SazbaDPH1", "12")
            self.add_element(output, "SazbaDPH2", "21")
            self.add_element(output, "Proplatit", invoice.invoice["czk_total_sum"])
            self.add_element(output, "Vyuctovano", "0")
            self.add_amounts(output, invoice, "czk_")
            if invoice.currency != "CZK":
                valuty = self.add_element(output, "Valuty")
                mena = self.add_element(valuty, "Mena")
                self.add_element(mena, "Kod", "EUR")
                self.add_element(mena, "Mnozstvi", "1")
                self.add_element(mena, "Kurs", invoice.invoice["czk_rate"])
                self.add_amounts(valuty, invoice)

            self.add_element(output, "PriUhrZbyv", "0")
            if invoice.currency != "CZK":
                self.add_element(output, "ValutyProp", invoice.invoice["total_sum"])
            self.add_element(output, "SumZaloha", "0")
            self.add_element(output, "SumZalohaC", "0")

            prijemce = self.add_element(output, "DodOdb")
            self.add_element(prijemce, "ObchNazev", invoice.contact["name"])
            adresa = self.add_element(prijemce, "ObchAdresa")
            self.add_element(adresa, "Ulice", invoice.contact["address"])
            self.add_element(adresa, "Misto", invoice.contact["city"])
            self.add_element(adresa, "Stat", invoice.contact["country"])
            self.add_element(prijemce, "FaktNazev", invoice.contact["name"])
            if invoice.contact["tax_reg"] and invoice.contact["vat_reg"].startswith(
                "CZ"
            ):
                self.add_element(prijemce, "ICO", invoice.contact["tax_reg"])
            if invoice.contact["vat_reg"]:
                self.add_element(prijemce, "DIC", invoice.contact["vat_reg"])
            adresa = self.add_element(prijemce, "FaktAdresa")
            self.add_element(adresa, "Ulice", invoice.contact["address"])
            self.add_element(adresa, "Misto", invoice.contact["city"])
            self.add_element(adresa, "Stat", invoice.contact["country"])
            if invoice.contact["vat_reg"]:
                self.add_element(prijemce, "PlatceDPH", "1")
                self.add_element(prijemce, "FyzOsoba", "0")

            seznam = self.add_element(output, "SeznamPolozek")
            for row in invoice.invoice["rows_data"]:
                polozka = self.add_element(seznam, "Polozka")
                self.add_element(polozka, "Popis", row["item"])
                self.add_element(polozka, "PocetMJ", row["quantity"])
                if invoice.currency == "CZK":
                    self.add_element(polozka, "Cena", row["total"])
                else:
                    self.add_element(polozka, "Valuty", row["total"])

        ElementTree.indent(document)
        ElementTree.dump(document)


@register_command
class Contacts(Command):
    """List invoices."""

    @classmethod
    def add_parser(cls, subparser):
        """Create parser for command line."""
        parser = super().add_parser(subparser)
        parser.add_argument(
            "--year",
            type=int,
            help="Year to process",
            default=datetime.date.today().year,
        )
        parser.add_argument(
            "--country",
            help="Country to list",
        )
        parser.add_argument("match", nargs="?", help="Match string to find")
        return parser

    def match(self, invoice):
        if self.args.country and self.args.country != invoice.contact["country"]:
            return False
        if not self.args.match:
            return True
        match = self.args.match.lower()
        return (
            match in invoice.invoice["item"].lower()
            or match in invoice.invoice["contact"].lower()
            or match in invoice.contact["name"].lower()
        )

    def run(self):
        """Execute the command."""
        contacts = {}
        for invoice in self.storage.list(self.args.year):
            if not self.match(invoice):
                continue
            contacts[invoice.contact["name"]] = invoice.contact

        for contact in contacts.values():
            print(f"{contact['name']}, {contact['city']}, {contact.get('email')}")


@register_command
class NotPaid(List):
    """Not paid invoices."""

    def match(self, invoice):
        return not invoice.paid() and super().match(invoice)


@register_command
class Detail(Command):
    """Show invoice detail."""

    @classmethod
    def add_parser(cls, subparser):
        """Create parser for command line."""
        parser = super().add_parser(subparser)
        parser.add_argument("id", help="Invoice id")
        return parser

    def run(self):
        """Execute the command."""
        invoice = self.storage.get(self.args.id)
        print(invoice.invoiceid)
        print("-" * len(invoice.invoiceid))
        print("Date:     ", invoice.invoice["date"])
        print("Due:      ", invoice.invoice["due"])
        print("Name:     ", invoice.contact["name"])
        print("Address:  ", invoice.contact["address"])
        print("City:     ", invoice.contact["city"])
        print("Country:  ", invoice.contact["country"])
        print("Item:     ", invoice.invoice["item"])
        print("Category: ", invoice.invoice["category"])
        print(f"Rate:      {invoice.rate} {invoice.currency}")
        print(f"Quantity:  {invoice.quantity}")
        print(f"Amount:    {invoice.amount} {invoice.currency}")
        print(f"Amount:    {invoice.amount_czk:.2f} CZK incl. VAT")
        if invoice.paid():
            print("Paid:      yes")
        else:
            print("Paid:      no")


@register_command
class WriteTex(Detail):
    """Generate tex."""

    def run(self):
        """Execute the command."""
        invoice = self.storage.get(self.args.id)
        invoice.write_tex()


@register_command
class BuildPDF(Detail):
    """Build PDF."""

    def run(self):
        """Execute the command."""
        invoice = self.storage.get(self.args.id)
        invoice.build_pdf()


@register_command
class Summary(Command):
    """Show invoice summary."""

    @classmethod
    def add_parser(cls, subparser):
        parser = super().add_parser(subparser)
        parser.add_argument(
            "--year",
            type=int,
            help="Year to process",
            default=datetime.date.today().year,
        )
        parser.add_argument(
            "--vat",
            action="store_true",
            help="Include VAT",
            default=False,
        )
        parser.add_argument("--summary", "-s", action="store_true", help="show YTD sum")
        return parser

    def run(self):
        categories = self.storage.settings["categories"].split(",")
        supertotal = 0
        year = self.args.year
        supercats = dict.fromkeys(categories, 0)
        cat_format = " ".join(f"{{{x}:7.0f}} CZK" for x in categories)
        header = "Month         Total {}".format(
            " ".join(f"{x.title():>11}" for x in categories),
        )
        print(header)
        print("-" * len(header))
        for month in range(1, 13):
            total = 0
            cats = dict.fromkeys(categories, 0)
            for invoice in self.storage.list(year, month):
                amount = invoice.amount_czk_vat if self.args.vat else invoice.amount_czk
                cats[invoice.category] += amount
                supercats[invoice.category] += amount
                total += amount
                supertotal += amount
            if self.args.summary:
                display_total = supertotal
                cat_sums = cat_format.format(**supercats)
            else:
                display_total = total
                cat_sums = cat_format.format(**cats)
            print(f"{year}/{month:02d} {display_total:7.0f} CZK {cat_sums}")
        print("-" * len(header))
        print(f"Summary {supertotal:7.0f} CZK {cat_format.format(**supercats)}")


@register_command
class Add(Command):
    """Create new invoice."""

    @classmethod
    def add_parser(cls, subparser):
        """Create parser for command line."""
        parser = super().add_parser(subparser)
        parser.add_argument(
            "--skip-validation",
            "-s",
            action="store_true",
            help="Skip VAT validation",
            default=False,
        )
        parser.add_argument("--edit", "-e", action="store_true", help="open in editor")
        parser.add_argument("contact", help="Contact name")
        return parser

    def run(self):
        contact = self.storage.read_contact(self.args.contact)
        vat_reg = contact.get("vat_reg", "")
        if vat_reg:
            vat_reg = vat_reg.strip().replace(" ", "")
            vatin = VATIN(vat_reg[:2], vat_reg[2:])
            if self.args.skip_validation:
                vatin.verify()
            elif not vatin.data.valid:
                raise ValueError(f"Invalid VAT: {vat_reg}")

        filename = self.storage.create(self.args.contact)
        print(filename)
        if self.args.edit:
            subprocess.run(["gvim", filename], check=True)


def main(args=None):
    """CLI entry point."""
    parser = ArgumentParser(
        description="Fakturace.",
        epilog="This utility is developed at <{}>.".format(
            "https://github.com/nijel/fakturace",
        ),
    )
    parser.add_argument(
        "--quotes",
        action="store_true",
        help="Operate on quotes instead of invoices",
    )
    parser.add_argument("--web", action="store_true", help="Operate on web invoices")
    parser.add_argument(
        "--proforma",
        action="store_true",
        help="Operate on proforma invoices",
    )

    subparser = parser.add_subparsers(dest="cmd")
    for command in COMMANDS:
        COMMANDS[command].add_parser(subparser)

    params = parser.parse_args(args)

    command = COMMANDS[params.cmd](params)
    command.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from setuptools import setup

with open("README.rst") as handle:
    LONG_DESCRIPTION = handle.read()

with open("requirements.txt") as handle:
    REQUIRES = handle.read().split()

setup(
    name="fakturace",
    version="0.7.1",
    author="Michal Čihař",
    author_email="michal@cihar.com",
    description="Simple invoicing and billing tool",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    license="GPLv3+",
    keywords="billing invoice",
    url="https://cihar.com/",
    download_url="https://github.com/nijel/fakturace",
    project_urls={"Issue Tracker": "https://github.com/nijel/fakturace/issues"},
    platforms=["any"],
    packages=["fakturace"],
    package_dir={"fakturace": "fakturace"},
    install_requires=REQUIRES,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["fakturace = fakturace.cli:main"]},
    python_requires=">=3.11",
)



import os
import subprocess
from configparser import RawConfigParser

from django.utils.functional import cached_property
from jinja2.exceptions import TemplateNotFound

from .data import CONTACT, DEFAULTS
from .rates import Rates


class Invoice:
    def __init__(self, storage, data, override=None):
        self.name = data
        self.storage = storage
        self.contact = {}
        self.invoice = {}
        self.bank = {}
        if override is None:
            self.override = {}
        else:
            self.override = override
        self.load()

    def load(self):
        """Load data from ini files."""
        data = RawConfigParser()
        data.read(self.name)

        self.invoice = dict(data["invoice"])

        self.contact = self.storage.read_contact(self.invoice["contact"])

        self.process_defaults()

        self.bank = self.storage.read_bank(
            self.invoice["currency"],
            self.invoice["bank_suffix"],
        )
        # Propagate defaults from bank
        for field in ("template", "note", "vat"):
            if field in self.bank:
                self.invoice[field] = self.bank[field]

        # Fetch invoice rows
        self.invoice["rows_data"] = []
        total_sum = 0
        for suffix in (
            "",
            "_2",
            "_3",
            "_4",
            "_5",
            "_6",
            "_7",
            "_8",
            "_9",
            "_10",
            "_11",
            "_12",
            "_13",
            "_14",
            "_15",
            "_16",
            "_17",
            "_18",
            "_19",
        ):
            if "item" + suffix not in self.invoice:
                break
            quantity = self.invoice.get("quantity" + suffix, "1")
            try:
                rate = float(self.invoice["rate" + suffix])
            except KeyError:
                rate = float(self.invoice["rate"])
            total = float(quantity.split()[0]) * rate
            total_sum += total
            self.invoice["rows_data"].append(
                {
                    "item": self.invoice["item" + suffix],
                    "rate": f"{rate:.2f}",
                    "quantity": quantity,
                    "total": f"{total:.2f}",
                    "currency": self.bank.get("currency", self.invoice["currency"]),
                },
            )

        self.invoice["total"] = f"{total_sum:.2f}"

        # Calculate VAT
        vat = int(self.invoice["vat"]) * total_sum / 100
        self.invoice["total_vat"] = f"{vat:.2f}"
        self.invoice["total_sum"] = f"{total_sum + vat:.2f}"
        self.invoice["include_czk"] = bool(int(self.invoice["vat"])) and self.contact[
            "vat_reg"
        ].startswith("CZ")

        # Rates in CZK
        rate = Rates.get(self.invoice["date"], self.currency)
        self.invoice["czk_rate"] = str(rate)
        self.invoice["czk_total"] = f"{rate * total_sum:.2f}"
        self.invoice["czk_total_vat"] = f"{rate * vat:.2f}"
        self.invoice["czk_total_sum"] = f"{rate * (total_sum + vat):.2f}"

        # Shorter summary for PDF title
        self.invoice["shortitem"] = self.invoice["item"].split(":")[0]

        remarks = []
        for pos in range(1, 10):
            name = f"remark_{pos}"
            if name not in self.invoice:
                break
            remarks.append(self.invoice[name])
        self.invoice["remarks"] = remarks

    def process_defaults(self):
        """Fill in default values."""
        for key, value in self.contact.items():
            if not key.startswith("default_"):
                continue
            name = key[8:]
            if name in self.invoice:
                continue
            self.invoice[name] = value
        for key, value in self.override.items():
            if key not in self.invoice:
                self.invoice[key] = value
        for key, value in DEFAULTS.items():
            if key not in self.invoice:
                self.invoice[key] = value
        for key, value in CONTACT.items():
            if key not in self.contact:
                self.contact[key] = value

    @cached_property
    def invoiceid(self):
        return os.path.splitext(os.path.basename(self.name))[0]

    @cached_property
    def tex_path(self):
        return self.storage.path(self.storage.tex, f"{self.invoiceid}.tex")

    @cached_property
    def pdf_path(self):
        return self.storage.path(self.storage.pdf, f"{self.invoiceid}.pdf")

    def write_tex(self):
        row_template = self.storage.jinja.get_template(self.invoice["row"])

        category_template = self.invoice["template"].replace(
            ".tex",
            f"-{self.category}.tex",
        )

        try:
            template = self.storage.jinja.get_template(category_template)
        except TemplateNotFound:
            template = self.storage.jinja.get_template(self.invoice["template"])

        rows = [row_template.render(row) for row in self.invoice["rows_data"]]

        context = {"invoiceid": self.invoiceid, "rows": "\n".join(rows)}
        context.update(self.contact)
        context.update(self.invoice)
        context.update(self.bank)
        output = template.render(context)
        self.storage.ensure_dir(self.tex_path)
        with open(self.tex_path, "w") as handle:
            handle.write(output)

    def build_pdf(self):
        self.storage.ensure_dir(self.pdf_path)
        subprocess.run(
            ["xelatex", os.path.abspath(self.tex_path)],
            check=True,
            cwd=self.storage.path(self.storage.pdf),
        )

    @property
    def category(self):
        return self.invoice["category"]

    @property
    def amount(self):
        return self.invoice["total"]

    @property
    def total_amount(self):
        if "total_sum" in self.invoice:
            return self.invoice["total_sum"]
        return self.invoice["total"]

    @property
    def currency(self):
        return self.invoice["currency"].split("-")[0]

    @property
    def rate(self):
        return self.invoice["rate"]

    @property
    def quantity(self):
        return self.invoice["quantity"]

    @property
    def amount_czk(self):
        rate = Rates.get(self.invoice["date"], self.currency)
        return float(self.invoice["total"]) * rate

    @property
    def amount_czk_vat(self):
        rate = Rates.get(self.invoice["date"], self.currency)
        if "total_sum" in self.invoice:
            total = self.invoice["total_sum"]
        else:
            total = self.invoice["total"]
        return float(total) * rate

    def paid(self):
        return os.path.exists(self.name.replace(".ini", ".paid"))

    @cached_property
    def paid_path(self):
        return self.storage.path(self.storage.data, f"{self.invoiceid}.paid")

    def mark_paid(self, text):
        with open(self.paid_path, "w") as handle:
            handle.write(text)


class Quote(Invoice):
    def __init__(self, storage, data):
        super().__init__(
            storage,
            data,
            {
                "template": "template/quote.tex",
                "note": (
                    "Should you have any questions concerning this quotation, "
                    "please contact us at sales@weblate.org."
                ),
            },
        )


class Proforma(Invoice):
    def __init__(self, storage, data):
        super().__init__(
            storage,
            data,
            {
                "template": "template/proforma.tex",
                "note": (
                    "This is not a tax invoice, "
                    "you will receive proper invoice upon payment."
                ),
            },
        )

    def process_defaults(self):
        super().process_defaults()
        self.invoice["bank_suffix"] = "proforma"
